# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""A module that provides high-level tools for public use."""
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "create_app",
    "raise_if_migration_version_mismatch",
    "migration",
    "migration_with_rule",
)

import collections.abc
import os
import pathlib
import types
import typing

from mongorunway.application import applications
from mongorunway.application import config
from mongorunway.application import use_cases
from mongorunway.domain import migration as domain_migration
from mongorunway.domain import migration_business_rule as domain_rule

if typing.TYPE_CHECKING:
    from mongorunway.application.applications import MigrationApp

_ProcessT = typing.TypeVar("_ProcessT", bound=domain_migration.MigrationProcess)


@typing.overload
def create_app(
    name: str,
    configuration: typing.Optional[
        typing.Union[os.PathLike[str], str, pathlib.Path, config.Config]
    ] = None,
    *,
    raise_on_none: typing.Literal[True] = True,
    verbose_exc: bool = False,
) -> MigrationApp:
    ...


@typing.overload
def create_app(
    name: str,
    configuration: typing.Optional[
        typing.Union[os.PathLike[str], str, pathlib.Path, config.Config]
    ] = None,
    *,
    raise_on_none: typing.Literal[False] = False,
    verbose_exc: bool = False,
) -> typing.Optional[MigrationApp]:
    ...


def create_app(
    name: str,
    configuration: typing.Optional[
        typing.Union[os.PathLike[str], str, pathlib.Path, config.Config]
    ] = None,
    *,
    raise_on_none: bool = False,
    verbose_exc: bool = False,
) -> typing.Union[MigrationApp, typing.Optional[MigrationApp]]:
    r"""Creates a migration application.

    Creates a migration application based on the provided parameters.
    Uses a use case to initialize the application configuration.

    Parameters
    ----------
    name : str
        The name of the application for which the creation options will
        be taken from the configuration file. The application name should
        match the application name in the configuration file.
    configuration : typing.Optional[typing.Union[os.PathLike[str], str, pathlib.Path, config.Config]]
        The path to the configuration file or the configuration object
        required for initializing the application.
    raise_on_none : bool, optional
        By default, the use-cases return None and do not raise exceptions.
        If `raise_on_none` is True, an exception will be raised on a failed
        attempt to initialize the application.
    verbose_exc : bool, optional
        In case of an exception being raised on a failed attempt to initialize
        the application, the exception information will be more detailed if
        `verbose_exc` is True.

    Raises
    ------
    RuntimeError
        If `raise_on_none` is True and the initialization of the application fails,
        this exception will be raised.

    Returns
    -------
    typing.Union[MigrationApp, typing.Optional[MigrationApp]]
        If `raise_on_none` is True, an application object is guaranteed to be
        returned or an exception is raised. Otherwise, if the application
        initialization fails, None is returned.
    """

    if not isinstance(configuration, config.Config) or configuration is None:
        configuration = use_cases.read_configuration(
            config_filepath=str(configuration) if configuration is not None else configuration,
            app_name=name,
            verbose_exc=verbose_exc,
        )

    if configuration is not use_cases.UseCaseFailed:
        return applications.MigrationAppImpl(configuration=configuration)

    if raise_on_none:
        raise RuntimeError(f"Creation of {name!r} application is failed.")

    return None


def raise_if_migration_version_mismatch(
    application: applications.MigrationApp,
    expected_version: typing.Union[int, typing.Callable[[], int]],
) -> None:
    r"""Raises an error if the versions do not match.

    Raises an error if the provided version or version getter does not
    match the version of the given migration application. The expected
    version or the return value of the expected version getter should
    be an instance of the `int` class. Otherwise, the comparison operation
    may behave unpredictably.

    Parameters
    ----------
    application : applications.MigrationApp
        The migration application whose version you want to check.
    expected_version : typing.Union[int, typing.Callable[[], int]]
        The expected version of the current migration application.
        The expected version or the return value of the expected version
        getter should be an instance of the `int` class. Otherwise, the
        comparison operation may behave unpredictably.

    Raises
    ------
    ValueError
        Raises an error if the provided version or version getter does
        not match the version of the given migration application.
    """

    if callable(expected_version):
        expected_version = expected_version()

    if (current_version := (application.session.get_current_version() or 0)) != expected_version:
        raise ValueError(
            f"Migration version mismatch."
            " "
            f"Actual: {current_version!r}, but {expected_version!r} expected."
        )

    return None


def migration(process_func: types.FunctionType, /) -> domain_migration.MigrationProcess:
    r"""Wraps a function in a migration process.

    Wraps a function in a migration process. The function should return
    either a ready-to-use migration process object or a sequence containing
    implementations of migration command interfaces. Otherwise, an exception
    will be raised (see the `Raises` section).

    !!! note
        **__name__**: If the provided object does not have the `__name__`
        attribute, the default name `UNDEFINED_PROCESS` will be set, which is
        used for debugging and logging migration processes.

        **__globals__**: If the provided object does not have the `__globals__`
        attribute, an `AttributeError` will be raised (see the `Raises`
        section).

    Parameters
    ----------
    process_func : types.FunctionType
        The function that returns either a ready-to-use migration process object
        or a sequence containing implementations of migration command interfaces.

    Returns
    -------
    domain_migration.MigrationProcess
        A migration process object that contains information about the migration
        version, process name, and set of migration commands. If the function
        returns an already created migration process, no new objects will be
        created.

    Raises
    ------
    ValueError
        If the function did not return an instance of the migration process
        class and the return value is not an instance of collections.abc.Sequence.
    AttributeError
        If the function did not return an instance of the migration process
        class and the module where the function is implemented does not contain
        the global value `version`.
    """

    func_callback = process_func()
    if isinstance(func_callback, domain_migration.MigrationProcess):
        return func_callback

    if not isinstance(func_callback, collections.abc.Sequence):
        raise ValueError(
            f"Migration process func {process_func!r} must return sequence of commands."
        )

    version = getattr(process_func, "__globals__", {}).get("version", None)
    if version is None:
        func_file = ""
        if hasattr(process_func, "__code__"):
            func_file = process_func.__code__.co_filename

        raise AttributeError(f"Migration module {func_file!r} should have 'version' variable.")

    return domain_migration.MigrationProcess(
        func_callback,
        migration_version=version,
        name=getattr(process_func, "__name__", "UNDEFINED_PROCESS"),
    )


def migration_with_rule(
    rule: domain_rule.MigrationBusinessRule, /
) -> typing.Callable[[_ProcessT], _ProcessT]:
    r"""Adds a rule to the migration process.

    Returns the provided migration process object with the added rule.

    Parameters
    ----------
    rule : domain_rule.MigrationBusinessRule
        The rule to add to the migration process.

    Raises
    ------
    ValueError
        If the value passed to the decorator is not an instance of the migration
        process class.
    """

    if not isinstance(rule, domain_rule.MigrationBusinessRule):
        raise ValueError(f"Rule must be instance of {domain_rule.MigrationBusinessRule!r}.")

    def decorator(process: _ProcessT) -> _ProcessT:
        process.add_rule(rule)
        return process

    return decorator



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "read",
    "getcwd",
    "join",
    "find_any",
)

import os
import typing


def read(filename: str) -> str:
    with open(filename) as file:
        return file.read()


def getcwd() -> str:
    return os.getcwd()


def join(*components: str) -> str:
    return os.path.join(*components)


def find_any(*filenames: str, base_dir: str = os.getcwd()) -> typing.List[str]:
    found_files: typing.List[str] = []

    for dirpath, dirnames, filenames_in_dir in os.walk(base_dir):
        for filename in filenames:
            if filename in filenames_in_dir:
                found_files.append(os.path.join(dirpath, filename))

    return found_files



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "INFO",
    "ERROR",
    "SUCCESS",
    "WARNING",
    "HEADING_MAP",
    "HEADING_LEVEL_ONE",
    "HEADING_LEVEL_TWO",
    "HEADING_LEVEL_THREE",
    "TOOL_HEADING_NAME",
    "AsciiOutput",
    "print",
    "print_heading",
    "print_error",
    "print_success",
    "print_info",
    "print_warning",
    "print_new_line",
    "verbose_print",
)

import enum
import typing

# In a public tool, dependency on CLI is an essential part, so in this case,
# we can ignore the Dependency Inversion Principle and implement output with
# a specific implementation using click.
import click

INFO = "info"
ERROR = "error"
SUCCESS = "success"
WARNING = "warning"

HEADING_LEVEL_ONE = 1
HEADING_LEVEL_TWO = 2
HEADING_LEVEL_THREE = 3

HEADING_MAP = {
    HEADING_LEVEL_ONE: ("=", True),
    HEADING_LEVEL_TWO: ("-", True),
    HEADING_LEVEL_THREE: ("-", False),
}

TOOL_HEADING_NAME: typing.Final[str] = "Mongorunway"


class AsciiOutput(str, enum.Enum):
    INFO = 128712
    DONE = 10003
    WARNING = 9888
    ERROR = 33

    def __str__(self) -> str:
        return chr(self.value)


def print(
    text: str = "",
    bold: bool = False,
    newline: bool = True,
    symbol: str = "",
) -> None:
    if symbol:
        text = symbol + " " + text

    click.secho(text, bold=bold, nl=newline)


def verbose_print(
    verbose: bool,
    text: str = "",
    bold: bool = False,
    newline: bool = True,
    symbol: str = "",
) -> None:
    if verbose:
        print(
            text=text,
            bold=bold,
            symbol=symbol,
            newline=newline,
        )


def print_new_line(count: int = 1) -> None:
    for i in range(count):
        click.secho()


def print_heading(level: int, text: str, indent: bool = True) -> None:
    line_char, show_line_above = HEADING_MAP[level]
    heading_line = line_char * len(text)

    if show_line_above:
        print(heading_line, bold=True)

    print(text, bold=True)
    print(heading_line, bold=True)

    if indent:
        print_new_line()


def print_success(text: str, bold: bool = False) -> None:
    print(text, bold=bold, symbol=AsciiOutput.DONE)


def print_error(text: str, bold: bool = False) -> None:
    print(text, bold=bold, symbol=AsciiOutput.ERROR)


def print_warning(text: str, bold: bool = False) -> None:
    print(text, bold=bold, symbol=AsciiOutput.WARNING)


def print_info(text: str, bold: bool = False) -> None:
    print(text, bold=bold, symbol=AsciiOutput.INFO)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "Migration",
    "MigrationProcess",
    "MigrationReadModel",
)

import dataclasses
import typing

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration_business_rule as domain_rule
    from mongorunway.domain import migration_command as domain_command

_ProcessT = typing.TypeVar("_ProcessT", bound="MigrationProcess")


class Migration:
    __slots__: typing.Sequence[str] = (
        "_name",
        "_version",
        "_checksum",
        "_is_applied",
        "_description",
        "_upgrade_process",
        "_downgrade_process",
    )

    def __init__(
        self,
        *,
        name: str,
        version: int,
        checksum: str,
        is_applied: bool,
        description: str,
        upgrade_process: MigrationProcess,
        downgrade_process: MigrationProcess,
    ) -> None:
        self._name = name
        self._version = version
        self._checksum = checksum
        self._is_applied = is_applied
        self._description = description
        self._upgrade_process = upgrade_process
        self._downgrade_process = downgrade_process

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> int:
        return self._version

    @property
    def checksum(self) -> str:
        return self._checksum

    @property
    def description(self) -> str:
        return self._description

    @property
    def is_applied(self) -> bool:
        return self._is_applied

    @property
    def upgrade_process(self) -> MigrationProcess:
        return self._upgrade_process

    @property
    def downgrade_process(self) -> MigrationProcess:
        return self._downgrade_process

    def set_is_applied(self, value: bool, /) -> None:
        self._is_applied = value

    def to_dict(self, *, unique: bool = False) -> typing.Dict[str, typing.Any]:
        mapping = {
            "name": self.name,
            "version": self.version,
            "checksum": self.checksum,
            "is_applied": self.is_applied,
            "description": self.description,
        }

        if unique:
            mapping["_id"] = self.version

        return mapping


@dataclasses.dataclass
class MigrationReadModel:
    name: str
    version: int
    checksum: str
    description: str
    is_applied: bool

    @classmethod
    def from_dict(cls, mapping: typing.MutableMapping[str, typing.Any], /) -> MigrationReadModel:
        mapping.pop("_id", None)  # For mongo records
        return cls(**mapping)

    @classmethod
    def from_migration(cls, migration: Migration, /) -> MigrationReadModel:
        return cls(
            name=migration.name,
            version=migration.version,
            checksum=migration.checksum,
            description=migration.description,
            is_applied=migration.is_applied,
        )


class MigrationProcess:
    def __init__(
        self,
        commands: domain_command.AnyCommandSequence,
        migration_version: int,
        name: str,
    ) -> None:
        self._rules: domain_rule.RuleSequence = []
        self._name = name
        self._commands = commands
        self._migration_version = migration_version

    @property
    def name(self) -> str:
        return self._name

    @property
    def commands(self) -> domain_command.AnyCommandSequence:
        return self._commands

    @property
    def migration_version(self) -> int:
        return self._migration_version

    @property
    def rules(self) -> domain_rule.RuleSequence:
        return self._rules

    def has_rules(self) -> bool:
        return bool(self._rules)

    def add_rule(self: _ProcessT, rule: domain_rule.MigrationBusinessRule, /) -> _ProcessT:
        self._rules.append(rule)
        return self



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MigrationEventManager",
)

import abc
import typing

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration_event as domain_event


class MigrationEventManager(abc.ABC):
    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    def subscribe_events(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def unsubscribe_events(self, event: typing.Type[domain_event.MigrationEvent]) -> None:
        ...

    @abc.abstractmethod
    def subscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def unsubscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def get_event_handlers_for(
        self,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> typing.MutableSequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]]:
        ...

    @abc.abstractmethod
    def prioritize_handler(
        self,
        handler: domain_event.EventHandler,
        event: typing.Type[domain_event.MigrationEvent],
        priority: int,
    ) -> None:
        ...

    @abc.abstractmethod
    def unprioritize_handler_proxy(
        self,
        handler_proxy: domain_event.EventHandlerProxy,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def listen(
        self,
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> typing.Callable[
        [domain_event.EventHandlerProxyOr[domain_event.EventHandlerT]],
        domain_event.EventHandlerProxyOr[domain_event.EventHandlerT],
    ]:
        ...

    @abc.abstractmethod
    def dispatch(self, event: domain_event.MigrationEvent) -> None:
        ...



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "DocumentType",
    "Client",
    "Database",
    "Collection",
    "ClientSession",
    "Cursor",
    "hint_or_sort_cursor",
    "translate_index",
)

import typing

import pymongo
from pymongo import client_session
from pymongo import database

DocumentType: typing.TypeAlias = typing.Dict[str, typing.Any]

Client = pymongo.MongoClient[DocumentType]

Database = database.Database[DocumentType]

Collection = pymongo.collection.Collection[DocumentType]

ClientSession = client_session.ClientSession

Cursor = pymongo.cursor.Cursor[DocumentType]


def translate_index(
    indexes: typing.Union[str, typing.Sequence[typing.Tuple[str, int]], typing.Tuple[str, int]], /
) -> str:
    if isinstance(indexes, str):
        return indexes
    if isinstance(indexes, tuple) and isinstance(indexes[0], str):
        return f"{indexes[0]}_{indexes[1]}"

    translated_index = "_".join(f"{x}_{y}" for x, y in indexes)
    return translated_index


def hint_or_sort_cursor(
    cursor: Cursor,
    /,
    indexes: typing.Union[str, typing.Sequence[typing.Tuple[str, int]]],
) -> Cursor:
    index_info = cursor.collection.index_information()
    if isinstance(indexes, str):
        if indexes not in index_info:
            return cursor.sort(indexes)
    else:
        for index in indexes:
            if isinstance(index, tuple):
                index_name, _ = index
                if index_name not in index_info:
                    return cursor.sort(indexes)

            if isinstance(index, str):
                if index not in index_info:
                    return cursor.sort(indexes)

    return cursor.hint(indexes)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = ("AuditlogJournalImpl",)

import dataclasses
import datetime
import typing

import pymongo

from mongorunway.application.ports import auditlog_journal as auditlog_journal_port
from mongorunway.domain import migration as domain_migration
from mongorunway.domain import migration_auditlog_entry as domain_auditlog_entry

if typing.TYPE_CHECKING:
    from mongorunway import mongo


class AuditlogJournalImpl(auditlog_journal_port.AuditlogJournal):
    __slots__: typing.Sequence[str] = ("_collection", "_max_records")

    def __init__(
        self,
        auditlog_collection: mongo.Collection,
        max_records: typing.Optional[int] = None,
    ) -> None:
        self._max_records = max_records
        self._collection = auditlog_collection

    @property
    def max_records(self) -> typing.Optional[int]:
        return self._max_records

    def set_max_records(self, value: typing.Optional[int], /) -> None:
        self._max_records = value

    def append_entries(
        self,
        entries: typing.Sequence[domain_auditlog_entry.MigrationAuditlogEntry],
    ) -> None:
        total = self._collection.count_documents({})

        if self._max_records is not None:
            remove = max(0, total - self._max_records + len(entries))
            if remove:
                ids = [r["_id"] for r in self._collection.find().limit(remove)]

                # Delete extra records based on the FIFO algorithm.
                self._collection.delete_many({"_id": {"$in": ids}})

        self._collection.insert_many(
            [dataclasses.asdict(entry) for entry in entries],
            # Audit log records have an automatically generated
            # identifier that does not need to be sorted.
            ordered=False,
        )

    def load_entries(
        self, limit: typing.Optional[int] = None
    ) -> typing.Sequence[domain_auditlog_entry.MigrationAuditlogEntry]:
        pipeline: typing.List[typing.Any] = [{"$match": {}}]
        if limit is not None:
            pipeline.append({"$limit": limit})

        entries = [
            domain_auditlog_entry.MigrationAuditlogEntry.from_dict(entry)
            for entry in self._collection.aggregate(pipeline)
        ]

        return entries

    def history(
        self,
        start: typing.Optional[datetime.datetime] = None,
        end: typing.Optional[datetime.datetime] = None,
        limit: typing.Optional[int] = None,
        ascending_date: bool = True,
    ) -> typing.Iterator[domain_auditlog_entry.MigrationAuditlogEntry]:
        pipeline: typing.List[typing.Any] = [
            {"$sort": {"date": pymongo.ASCENDING if ascending_date else pymongo.DESCENDING}}
        ]
        if start is not None:
            pipeline.append({"$match": {"date": {"$gte": start}}})
        if end is not None:
            pipeline.append({"$match": {"date": {"$lte": end}}})
        if limit is not None:
            pipeline.append({"$limit": limit})

        schemas = self._collection.aggregate(pipeline)

        for schema in schemas:
            schema["migration_read_model"] = domain_migration.MigrationReadModel.from_dict(
                schema["migration_read_model"],
            )
            yield domain_auditlog_entry.MigrationAuditlogEntry.from_dict(schema)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MissingFilenameStrategy",
    "NumericalFilenameStrategy",
    "UnixFilenameStrategy",
)

import re
import time
import typing

from mongorunway.application.ports import filename_strategy as filename_strategy_port


class MissingFilenameStrategy(filename_strategy_port.FilenameStrategy):
    __slots__: typing.Sequence[str] = ()

    def is_valid_filename(self, filename: str, /) -> bool:
        return True

    def transform_migration_filename(self, filename: str, position: int) -> str:
        return filename


class NumericalFilenameStrategy(filename_strategy_port.FilenameStrategy):
    __slots__: typing.Sequence[str] = ()

    def is_valid_filename(self, filename: str, /) -> bool:
        return all(char.isdigit() for char in filename[:3])

    def transform_migration_filename(self, filename: str, position: int) -> str:
        if not self.is_valid_filename(filename):
            filename_parts = [i for i in filename.split("_") if i]
            return str(position).zfill(3) + "_" + "_".join(filename_parts)

        return filename


class UnixFilenameStrategy(filename_strategy_port.FilenameStrategy):
    __slots__: typing.Sequence[str] = ()

    def is_valid_filename(self, filename: str, /) -> bool:
        return bool(re.match(r"^\d{10,}", filename))

    def transform_migration_filename(self, filename: str, position: int) -> str:
        if not self.is_valid_filename(filename):
            filename_parts = [i for i in filename.split("_") if i]
            return str(int(time.time())) + "_" + "_".join(filename_parts)

        return filename



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""The transaction module provides classes for performing transactions in a migration application.
It contains implementations of the base transaction class as well as classes for different types of
transactions.
"""
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MigrationTransaction",
    "UpgradeTransaction",
    "DowngradeTransaction",
    "TRANSACTION_SUCCESS",
    "TRANSACTION_NOT_APPLIED",
)

import abc
import logging
import typing

from mongorunway import util
from mongorunway.application import session
from mongorunway.application.services import validation_service
from mongorunway.domain import migration_context as domain_context

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration as domain_migration

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("mongorunway.transactions")
_SelfT = typing.TypeVar("_SelfT", bound="MigrationTransaction")

TransactionCode: typing.TypeAlias = int

TRANSACTION_SUCCESS: typing.Final[TransactionCode] = 1

TRANSACTION_NOT_APPLIED: typing.Final[TransactionCode] = 0


class MigrationTransaction(abc.ABC):
    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def exc_val(self) -> typing.Optional[BaseException]:
        ...

    @classmethod
    @abc.abstractmethod
    def create(
        cls: typing.Type[_SelfT],
        migration_session: session.MigrationSession,
        migration: domain_migration.Migration,
    ) -> _SelfT:
        ...

    @abc.abstractmethod
    def is_failed(self) -> bool:
        ...

    @abc.abstractmethod
    def get_process(
        self, migration: domain_migration.Migration, /
    ) -> domain_migration.MigrationProcess:
        ...

    @abc.abstractmethod
    def apply_to(self, session_context: session.MongoSessionContext) -> None:
        ...

    @abc.abstractmethod
    def commit(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        ...

    @abc.abstractmethod
    def rollback(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        ...


class AbstractMigrationTransaction(MigrationTransaction, abc.ABC):
    def __init__(
        self,
        migration_session: session.MigrationSession,
        migration: domain_migration.Migration,
    ) -> None:
        self._client = migration_session.session_client
        self._migration_session = migration_session
        self._migration = migration
        self._exc_val: typing.Optional[BaseException] = None

    @classmethod
    def create(
        cls: typing.Type[_SelfT],
        migration_session: session.MigrationSession,
        migration: domain_migration.Migration,
    ) -> _SelfT:
        return cls(migration_session, migration)  # type: ignore[call-arg]

    @property
    def exc_val(self) -> typing.Optional[BaseException]:
        return self._exc_val

    def is_failed(self) -> bool:
        return self.exc_val is not None

    @typing.final
    def apply_to(self, session_context: session.MongoSessionContext) -> None:
        process = self.get_process(self._migration)
        context = self._build_command_context(session_context)
        validation_service.validate_migration_process(process, context)

        mongodb_session_id = util.hexlify(session_context.mongodb_session_id)
        try:
            waiting_commands_count = len(process.commands)
            with session_context.start_transaction():
                _LOGGER.info(
                    "Beginning a transaction in MongoDB session (%s) for (%s) process.",
                    mongodb_session_id,
                    process.name,
                )

                for command_idx, command in enumerate(process.commands, 1):
                    command.execute(context)

                    _LOGGER.info(
                        "%s command successfully applied (%s of %s).",
                        command.name,
                        command_idx,
                        waiting_commands_count,
                    )

                self.commit(self._migration, session_context)

        except Exception as exc:
            _LOGGER.error(
                "Transaction execution in MongoDB session (%s) ended with error %s.",
                mongodb_session_id,
                type(exc).__name__,
            )
            _LOGGER.debug("Error details of transaction execution: %s", str(exc))

            self._exc_val = exc
            self.rollback(self._migration, session_context)

        if session_context.has_ended:
            _LOGGER.info("MongoDB session %s has ended.", mongodb_session_id)

    @typing.final
    def rollback(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        _LOGGER.debug(
            "Rolling back migration %s with version %s",
            migration.name,
            migration.version,
        )
        self._rollback(migration, mongo_session)

    @typing.final
    def commit(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        _LOGGER.debug(
            "Committing migration %s with version %s",
            migration.name,
            migration.version,
        )
        self._commit(migration, mongo_session)

    @abc.abstractmethod
    def _rollback(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        pass

    @abc.abstractmethod
    def _commit(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        pass

    def _build_command_context(
        self,
        session_context: session.MongoSessionContext,
    ) -> domain_context.MigrationContext:
        return domain_context.MigrationContext(
            mongorunway_session_id=util.hexlify(self._migration_session.session_id),
            mongodb_session_id=util.hexlify(session_context.mongodb_session_id),
            client=self._migration_session.session_client,
            database=self._migration_session.session_database,
        )


class UpgradeTransaction(AbstractMigrationTransaction):
    def get_process(
        self, migration: domain_migration.Migration, /
    ) -> domain_migration.MigrationProcess:
        return migration.upgrade_process

    def _rollback(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        self._migration_session.set_applied_flag(migration, False)

    def _commit(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        self._migration_session.set_applied_flag(migration, True)


class DowngradeTransaction(AbstractMigrationTransaction):
    def get_process(
        self, migration: domain_migration.Migration, /
    ) -> domain_migration.MigrationProcess:
        return migration.downgrade_process

    def _rollback(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        self._migration_session.set_applied_flag(migration, True)

    def _commit(
        self,
        migration: domain_migration.Migration,
        mongo_session: session.MongoSessionContext,
    ) -> None:
        self._migration_session.set_applied_flag(migration, False)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = ("AuditlogJournal",)

import abc
import datetime
import typing

if typing.TYPE_CHECKING:
    from mongorunway.domain import migration_auditlog_entry as domain_auditlog_entry


class AuditlogJournal(abc.ABC):
    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def max_records(self) -> typing.Optional[int]:
        ...

    @abc.abstractmethod
    def set_max_records(self, value: typing.Optional[int], /) -> None:
        ...

    @abc.abstractmethod
    def append_entries(
        self,
        entries: typing.Sequence[domain_auditlog_entry.MigrationAuditlogEntry],
    ) -> None:
        ...

    @abc.abstractmethod
    def load_entries(
        self, limit: typing.Optional[int] = None
    ) -> typing.Sequence[domain_auditlog_entry.MigrationAuditlogEntry]:
        ...

    @abc.abstractmethod
    def history(
        self,
        start: typing.Optional[datetime.datetime] = None,
        end: typing.Optional[datetime.datetime] = None,
        limit: typing.Optional[int] = None,
        ascending_date: bool = True,
    ) -> typing.Iterator[domain_auditlog_entry.MigrationAuditlogEntry]:
        ...



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "default_repository_reader",
    "default_auditlog_journal_reader",
    "read_repository",
    "read_events",
    "read_event_handlers",
    "read_auditlog_journal",
    "read_filename_strategy",
    "BaseConfigReader",
    "YamlConfigReader",
)

import abc
import re
import typing

import yaml  # type: ignore[import]

from mongorunway import mongo
from mongorunway import util
from mongorunway.application import config
from mongorunway.application import filesystem
from mongorunway.application.ports import auditlog_journal as auditlog_journal_port
from mongorunway.application.ports import config_reader as config_reader_port
from mongorunway.application.ports import filename_strategy as filename_strategy_port
from mongorunway.application.ports import repository as repository_port
from mongorunway.domain import migration_event as domain_event
from mongorunway.infrastructure.persistence import auditlog_journals
from mongorunway.infrastructure.persistence import repositories

event_handler_pattern: typing.Pattern[str] = re.compile(
    r"""
    ^                  # Start of the string
    Prioritized\[      # Match the literal string "Prioritized["
        (\d+),         # Group 1: Match one or more digits (priority)
        \s*            # Match zero or more whitespace characters
        (.+)           # Group 2: Match one or more characters (handler function path)
    \]$                # Match the closing square bracket and end of the string
    |                  # OR
    ^                  # Start of the string
    ([^,\[\]]+)        # Group 3: Match any characters except comma, square brackets
    $                  # End of the string
    """,
    flags=re.VERBOSE,
)

logging_config: typing.Dict[str, typing.Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simpleFormatter": {
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    },
    "handlers": {
        "consoleHandler": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simpleFormatter",
        }
    },
    "loggers": {"root": {"level": "INFO", "handlers": ["consoleHandler"], "propagate": 0}},
}


def default_repository_reader(
    application_data: typing.Dict[str, typing.Any],
) -> repository_port.MigrationModelRepository:
    client = mongo.Client(**util.build_mapping_values(application_data["app_client"]))
    database = client.get_database(application_data["app_database"])
    collection = database.get_collection(application_data["app_repository"]["collection"])
    return repositories.MongoModelRepositoryImpl(collection)


def default_auditlog_journal_reader(
    application_data: typing.Dict[str, typing.Any],
) -> typing.Optional[auditlog_journal_port.AuditlogJournal]:
    if (collection := application_data["app_auditlog_journal"].get("collection")) is None:
        return None

    client = mongo.Client(**util.build_mapping_values(application_data["app_client"]))
    database = client.get_database(application_data["app_database"])
    collection = database.get_collection(collection)
    return auditlog_journals.AuditlogJournalImpl(collection)


@typing.no_type_check
def read_event_handlers(
    handler_name_seq: typing.Sequence[str],
) -> typing.List[domain_event.EventHandlerProxyOr[domain_event.EventHandler]]:
    handlers: typing.List[domain_event.EventHandlerProxyOr[domain_event.EventHandler]] = []

    for handler_name in handler_name_seq:
        match = re.match(event_handler_pattern, handler_name.replace(" ", ""))
        if match:
            if match.group(1):  # If there is a group 1, then there is a structure "Priority".
                priority = int(match.group(1))
                handler_func_path = match.group(2)
            else:
                priority = None
                handler_func_path = match.group(3)

            try:
                handler = util.import_obj(handler_func_path, cast=domain_event.EventHandler)
            except AttributeError as exc:
                raise AttributeError(f"Undefined event handler: {handler_func_path!r}.") from exc

            if priority is not None:
                handler = domain_event.EventHandlerProxy(priority=priority, handler=handler)

            handlers.append(handler)
        else:
            raise ValueError("Invalid handler format.")

    return handlers


@typing.no_type_check
def read_events(
    event_dict: typing.Dict[str, typing.Sequence[str]],
) -> typing.Mapping[
    typing.Type[domain_event.MigrationEvent],
    typing.Sequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]],
]:
    try:
        mapping = {
            util.import_obj(event_name, cast=domain_event.MigrationEvent): read_event_handlers(
                handler_name_seq
            )
            for event_name, handler_name_seq in event_dict.items()
        }
    except AttributeError as exc:
        raise AttributeError(f"Undefined event or event handler.") from exc

    return mapping


@typing.no_type_check
def read_filename_strategy(strategy_path: str) -> filename_strategy_port.FilenameStrategy:
    try:
        strategy_type = util.import_obj(
            strategy_path.strip(),
            cast=filename_strategy_port.FilenameStrategy,
        )
    except AttributeError as exc:
        raise AttributeError(f"Undefined filename strategy {strategy_path!r}.") from exc

    return strategy_type()


@typing.no_type_check
def read_repository(
    application_data: typing.Dict[str, typing.Any],
) -> repository_port.MigrationModelRepository:
    try:
        if (repository_data := application_data.get("app_repository")) is None:
            raise KeyError("Missing 'app_repository' section.")

        if (reader_value := repository_data.get("reader")) is not None:
            reader = util.import_obj(
                reader_value,
                cast=typing.Callable[
                    [typing.Dict[str, typing.Any]],
                    repository_port.MigrationModelRepository,
                ],
            )
            return reader(application_data)

        if (repo_type := application_data.get("type")) is None:
            reader = util.import_obj(
                "mongorunway.infrastructure.config_readers.default_repository_reader",
                cast=typing.Callable[
                    [typing.Dict[str, typing.Any]],
                    repository_port.MigrationModelRepository,
                ],
            )
            return reader(application_data)

        repository_type = util.import_obj(
            repo_type,
            cast=repository_port.MigrationModelRepository,
        )

    except AttributeError as exc:
        raise AttributeError(f"Undefined reader or repository received.") from exc

    return repository_type()


@typing.no_type_check
def read_auditlog_journal(
    application_data: typing.Dict[str, typing.Any],
) -> typing.Optional[auditlog_journal_port.AuditlogJournal]:
    try:
        if (auditlog_value := application_data.get("app_auditlog_journal")) is None:
            return None

        if (reader_value := auditlog_value.get("reader")) is not None:
            reader = util.import_obj(
                reader_value,
                cast=typing.Callable[
                    [typing.Dict[str, typing.Any]],
                    auditlog_journal_port.AuditlogJournal,
                ],
            )
            return reader(application_data)

        if (audit_type := application_data.get("type")) is None:
            reader = util.import_obj(
                "mongorunway.infrastructure.config_readers.default_auditlog_journal_reader",
                cast=typing.Callable[
                    [typing.Dict[str, typing.Any]],
                    auditlog_journal_port.AuditlogJournal,
                ],
            )
            return reader(application_data)

        auditlog_journal_type = util.import_obj(
            audit_type,
            cast=auditlog_journal_port.AuditlogJournal,
        )
    except AttributeError as exc:
        raise AttributeError(f"Undefined reader or repository received.") from exc

    return auditlog_journal_type()


class BaseConfigReader(config_reader_port.ConfigReader):
    potential_config_filenames: typing.List[str]

    def __init__(self, application_name: str) -> None:
        self.application_name = application_name

    @classmethod
    def from_application_name(cls, application_name: str, /) -> config_reader_port.ConfigReader:
        return cls(application_name=application_name)

    def read_config(
        self,
        config_filepath: typing.Optional[str] = None,
    ) -> typing.Optional[config.Config]:
        if config_filepath:
            config_filepaths = filesystem.find_any(config_filepath)
            if not config_filepaths:
                raise FileNotFoundError(f"Could not find {config_filepath!r} configuration file.")
        else:
            config_filepaths = filesystem.find_any(*self.potential_config_filenames)
            if not config_filepaths:
                return None

        # Search until at least one config is found
        for config_filepath in config_filepaths:
            configuration = self._read_config(config_filepath)
            if configuration:
                return configuration

        return None

    @abc.abstractmethod
    def _read_config(self, config_filepath: str) -> typing.Optional[config.Config]:
        ...


class YamlConfigReader(BaseConfigReader):
    potential_config_filenames = ["mongorunway.yaml"]

    def _read_config(self, config_filepath: str) -> typing.Optional[config.Config]:
        with open(config_filepath, "r") as config_file:
            configration_data = yaml.safe_load(config_file)["mongorunway"]

        return config.Config(
            filesystem=self._read_filesystem_config(
                configration_data["filesystem"],
                config_filepath,
            ),
            application=self._read_application_config(
                configration_data["applications"][self.application_name],
            ),
            logging_dict=configration_data.get("logging", logging_config),
        )

    def _read_filesystem_config(
        self,
        filesystem_data: typing.Dict[str, typing.Any],
        config_filepath: str,
    ) -> config.FileSystemConfig:
        return config.FileSystemConfig(
            config_dir=config_filepath,
            scripts_dir=filesystem_data["scripts_dir"],
            filename_strategy=read_filename_strategy(
                filesystem_data.get(
                    "filename_strategy",
                    "mongorunway.infrastructure.filename_strategies.NumericalFilenameStrategy",
                ),
            ),
            **util.build_optional_kwargs(
                ("use_filename_strategy",),
                filesystem_data,
            ),
        )

    def _read_application_config(
        self,
        application_data: typing.Dict[str, typing.Any],
    ) -> config.ApplicationConfig:
        return config.ApplicationConfig(
            app_name=self.application_name,
            app_client=(
                client := mongo.Client(**util.build_mapping_values(application_data["app_client"]))
            ),
            app_database=client.get_database(application_data["app_database"]),
            app_repository=read_repository(application_data),
            app_auditlog_journal=read_auditlog_journal(application_data),
            app_subscribed_events=read_events(application_data.get("app_subscribed_events", {})),
            **util.build_optional_kwargs(
                (
                    "app_timezone",
                    "app_date_format",
                    "app_auditlog_limit",
                    "use_logging",
                    "use_auditlog",
                    "use_indexing",
                    "use_schema_validation",
                    "raise_on_transaction_failure",
                ),
                application_data,
            ),
        )



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = ("ConfigReader",)

import abc
import typing

if typing.TYPE_CHECKING:
    from mongorunway.application import config


class ConfigReader(abc.ABC):
    __slots__ = ()

    @classmethod
    @abc.abstractmethod
    def from_application_name(cls, application_name: str, /) -> ConfigReader:
        ...

    @abc.abstractmethod
    def read_config(
        self,
        config_filename: typing.Optional[str] = None,
    ) -> typing.Optional[config.Config]:
        ...



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "ALL",
    "ExitCode",
    "FAILURE",
    "MINUS",
    "PLUS",
    "SUCCESS",
    "UseCaseFailed",
    "UseCaseFailedOr",
    "create_migration_file",
    "downgrade",
    "get_auditlog_entries",
    "get_pushed_version",
    "get_status",
    "read_configuration",
    "render_downgrade_results",
    "render_error",
    "render_upgrade_results",
    "upgrade",
    "usecase",
    "walk",
    "refresh",
)

import datetime
import functools
import sys
import traceback
import typing

import typing_extensions

from mongorunway import util
from mongorunway.application import output
from mongorunway.application import ux
from mongorunway.application.ports import config_reader as config_reader_port
from mongorunway.application.services import migration_service
from mongorunway.application.services import status_service

if typing.TYPE_CHECKING:
    from mongorunway.application import applications
    from mongorunway.application import config
    from mongorunway.domain import migration_auditlog_entry as domain_auditlog_entry

_T = typing.TypeVar("_T")
try:
    _P = typing.ParamSpec("_P")
except AttributeError:
    _P = typing_extensions.ParamSpec("_P")

PLUS: typing.Final[str] = "+"
MINUS: typing.Final[str] = "-"
ALL: typing.Final[str] = "all"

ExitCode: typing.TypeAlias = int

SUCCESS: typing.Final[ExitCode] = 0

FAILURE: typing.Final[ExitCode] = 1

UseCaseFailed: typing.TypeAlias = object()
r"""Use case callback sentinel.

Some use cases that are designed to retrieve specific data may, by design, 
return None. However, since errors at the use case level are not raised but 
handled and logged to the console, it has been decided to make it a general 
rule to return the UseCaseFailed sentinel in case of a failed use case.

Notes
-----
It is important to differentiate between ExitCode and this sentinel: 'ExitCode' 
is returned by commands that are meant to perform a specific action without being 
tied to a particular result. On the other hand, as mentioned earlier, UseCaseFailed 
is intended for the opposite scenario: for commands that are meant  to search and 
compute data and are expected to return a result.

See Also
--------
ExitCode
"""

if sys.version_info < (3, 11):
    UseCaseFailed: typing.TypeAlias = typing.Any

UseCaseFailedOr = typing.Union[_T, UseCaseFailed]


def usecase(
    *,
    has_verbose_exc: bool,
) -> typing.Callable[[typing.Callable[_P, _T]], typing.Callable[_P, ExitCode]]:
    def decorator(func: typing.Callable[_P, _T]) -> typing.Callable[_P, ExitCode]:
        @functools.wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> ExitCode:
            try:
                func(*args, **kwargs)
            except BaseException as exc:
                render_error(
                    exc,
                    verbose=typing.cast(
                        bool,
                        kwargs["verbose_exc"] if has_verbose_exc else False,
                    ),
                )
                return FAILURE

            return SUCCESS

        return wrapper

    return decorator


def query_usecase(
    *,
    has_verbose_exc: bool,
) -> typing.Callable[[typing.Callable[_P, _T]], typing.Callable[_P, UseCaseFailedOr[_T]]]:
    def decorator(func: typing.Callable[_P, _T]) -> typing.Callable[_P, UseCaseFailedOr[_T]]:
        @functools.wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> UseCaseFailedOr[_T]:
            try:
                callback = func(*args, **kwargs)
            except BaseException as exc:
                render_error(
                    exc,
                    verbose=typing.cast(
                        bool,
                        kwargs["verbose_exc"] if has_verbose_exc else False,
                    ),
                )
                return UseCaseFailed

            return callback

        return wrapper

    return decorator


@usecase(has_verbose_exc=True)
def downgrade(
    application: applications.MigrationApp,
    expression: str,
    verbose: bool,
    verbose_exc: bool,
) -> ExitCode:
    func: typing.Optional[typing.Callable[..., ExitCode]] = None
    args: typing.Tuple[typing.Any, ...] = ()

    if expression.isdigit():
        func, args = application.downgrade_to, (int(expression),)
    elif expression == MINUS:
        func = application.downgrade_once
    elif len(expression) > 1 and expression.startswith(MINUS):
        func, args = application.downgrade_to, (
            int(expression) + (application.session.get_current_version() or 0),
        )
    elif expression == ALL:
        func = application.downgrade_all

    if func is None:
        raise ValueError(f"The following expression cannot be applied: {expression!r}")

    render_downgrade_results(
        verbose,
        *util.timeit_func(func, *args),
    )

    return SUCCESS


@usecase(has_verbose_exc=True)
def upgrade(
    application: applications.MigrationApp,
    expression: str,
    verbose: bool,
    verbose_exc: bool,
) -> ExitCode:
    func: typing.Optional[typing.Callable[..., ExitCode]] = None
    args: typing.Tuple[typing.Any, ...] = ()

    if expression.isdigit():
        func, args = application.upgrade_to, (int(expression),)
    elif expression == PLUS:
        func = application.upgrade_once
    elif expression.startswith(PLUS):
        func, args = application.upgrade_to, (
            int(expression[1:]) + (application.session.get_current_version() or 0),
        )
    elif expression == ALL:
        func = application.upgrade_all

    if func is None:
        raise ValueError(f"The following expression cannot be applied: {expression!r}")

    render_upgrade_results(
        verbose,
        *util.timeit_func(func, *args),
    )

    return SUCCESS


@usecase(has_verbose_exc=True)
def walk(
    application: applications.MigrationApp,
    expression: str,
    verbose: bool,
    verbose_exc: bool,
) -> ExitCode:
    if expression[0] not in {PLUS, MINUS}:
        raise ValueError(
            "This command can only go in positive or negative order. "
            "Therefore, the expression must begin with either the '+' "
            "or '-' character."
        )

    if expression.startswith(MINUS):
        return downgrade(
            application=application,
            expression=expression,
            verbose=verbose,
            verbose_exc=verbose_exc,
        )

    return upgrade(
        application=application,
        expression=expression,
        verbose=verbose,
        verbose_exc=verbose_exc,
    )


@usecase(has_verbose_exc=True)
def create_migration_file(
    application: applications.MigrationApp,
    migration_filename: str,
    verbose_exc: bool,
    migration_version: typing.Optional[int] = None,
) -> ExitCode:
    if migration_version is None:
        output.print_info("Migration version is not specified, using auto-incrementation...")

    service = migration_service.MigrationService(application.session)
    service.create_migration_file_template(migration_filename, migration_version)

    return SUCCESS


@usecase(has_verbose_exc=True)
def init(
    application: applications.MigrationApp,
    verbose_exc: bool,
    init_scripts_dir: bool,
    init_collection: bool,
    init_collection_indexes: bool,
    init_collection_schema_validation: bool,
) -> ExitCode:
    if init_scripts_dir:
        ux.configure_migration_directory(application.session.session_scripts_dir)
    if init_collection:
        ux.configure_migration_collection(
            application.session.session_database,
            use_schema_validation=init_collection_schema_validation,
        )
    if init_collection_indexes:
        ux.configure_migration_indexes(application.session.session_database.migrations)

    return SUCCESS


@usecase(has_verbose_exc=True)
def refresh(
    application: applications.MigrationApp,
    verbose_exc: bool,
) -> ExitCode:
    synced_names = ux.sync_scripts_with_repository(application)
    output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
    if synced_names:
        output.print_info(f"'{', '.join(synced_names)}' migrations was successfully synced.")
    else:
        output.print_info("There is no unsynced migrations.")

    return SUCCESS


@query_usecase(has_verbose_exc=True)
def get_auditlog_entries(
    application: applications.MigrationApp,
    verbose_exc: bool,
    start: typing.Optional[datetime.datetime] = None,
    end: typing.Optional[datetime.datetime] = None,
    limit: typing.Optional[int] = None,
    ascending_date: bool = True,
) -> UseCaseFailedOr[typing.Sequence[domain_auditlog_entry.MigrationAuditlogEntry]]:
    history = application.session.history(
        start=start,
        end=end,
        limit=limit,
        ascending_date=ascending_date,
    )

    return tuple(history)


@query_usecase(has_verbose_exc=True)
def get_status(
    application: applications.MigrationApp,
    verbose_exc: bool,
    pushed_depth: int = -1,
) -> UseCaseFailedOr[typing.Tuple[bool, int]]:
    all_pushed_successfully = status_service.check_if_all_pushed_successfully(
        application=application,
        depth=pushed_depth,
    )

    return all_pushed_successfully, pushed_depth


@query_usecase(has_verbose_exc=True)
def get_pushed_version(
    application: applications.MigrationApp,
    verbose_exc: bool,
) -> UseCaseFailedOr[typing.Optional[int]]:
    version = application.session.get_current_version()
    return version


@query_usecase(has_verbose_exc=True)
def read_configuration(
    config_filepath: typing.Optional[str],
    *,
    app_name: str,
    verbose_exc: bool,
) -> UseCaseFailedOr[config.Config]:
    reader: typing.Optional[config_reader_port.ConfigReader] = None
    if config_filepath is None or config_filepath.endswith(".yaml"):  # Default reader
        reader = util.import_obj(
            "mongorunway.infrastructure.config_readers.YamlConfigReader",
            cast=config_reader_port.ConfigReader,
        ).from_application_name(app_name)

    if reader is None:
        output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
        output.print_error("Undefined configuration file type.")
        return UseCaseFailed

    configuration = reader.read_config(config_filepath)
    if configuration is None:
        output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
        output.print_error(f"Cannot find any configuration files in {config_filepath} directory.")
        return UseCaseFailed

    return configuration


def render_error(exc: BaseException, verbose: bool = False) -> None:
    output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
    output.print_warning(type(exc).__name__ + " : " + str(exc))

    if verbose:
        exc_info = traceback.format_exception(*sys.exc_info())
        output.print_error("\n".join(exc_info))


def render_downgrade_results(verbose: bool, downgraded_count: int, executed_in: float) -> None:
    output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
    output.verbose_print(verbose, "Verbose mode enabled.")
    output.print_success(f"Successfully downgraded {downgraded_count} migration(s).")
    output.verbose_print(
        verbose,
        f"Downgraded {downgraded_count} migration(s) in {executed_in}s.",
    )


def render_upgrade_results(verbose: bool, upgraded_count: int, executed_in: float) -> None:
    output.print_heading(output.HEADING_LEVEL_ONE, output.TOOL_HEADING_NAME)
    output.verbose_print(verbose, "Verbose mode enabled.")
    output.print_success(f"Successfully upgraded {upgraded_count} migration(s).")
    output.verbose_print(verbose, f"Upgraded {upgraded_count} migration(s) in {executed_in}s.")



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MigrationApp",
    "MigrationAppImpl",
)

import abc
import functools
import logging
import typing

import typing_extensions

from mongorunway.application import event_manager
from mongorunway.application import session
from mongorunway.application import traits
from mongorunway.application import transactions
from mongorunway.application import ux
from mongorunway.application.services import migration_service
from mongorunway.application.services import versioning_service
from mongorunway.domain import migration as domain_migration
from mongorunway.domain import migration_event as domain_event
from mongorunway.domain import migration_exception as domain_exception

if typing.TYPE_CHECKING:
    from mongorunway.application import config
    from mongorunway.domain import migration_event_manager as domain_event_manager

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("mongorunway.ui")
try:
    _P = typing.ParamSpec("_P")
except AttributeError:
    _P = typing_extensions.ParamSpec("_P")
_TransactionCodeT = typing.TypeVar("_TransactionCodeT", bound=transactions.TransactionCode)


def requires_migrations(
    *,
    is_applied: bool,
) -> typing.Callable[
    [typing.Callable[_P, _TransactionCodeT]], typing.Callable[_P, _TransactionCodeT]
]:
    def decorator(
        meth: typing.Callable[_P, _TransactionCodeT],
    ) -> typing.Callable[_P, _TransactionCodeT]:
        @functools.wraps(meth)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _TransactionCodeT:
            if not isinstance((self := args[0]), traits.MigrationSessionAware):
                raise ValueError(
                    f"'requires_migrations' can be applied only to "
                    f"{traits.MigrationSessionAware!r} objects."
                )

            models = self.session.get_migration_models_by_flag(is_applied=is_applied)
            if not models:
                if self.session.raises_on_transaction_failure:
                    if is_applied:
                        raise domain_exception.NothingToDowngradeError()
                    raise domain_exception.NothingToUpgradeError()

                return typing.cast(
                    _TransactionCodeT,
                    transactions.TRANSACTION_NOT_APPLIED,
                )
            return meth(*args, **kwargs)

        return wrapper

    return decorator


class MigrationApp(
    traits.MigrationRunner,
    traits.MigrationSessionAware,
    traits.MigrationEventManagerAware,
    abc.ABC,
):
    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...

    @abc.abstractmethod
    def subscribe_events(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def unsubscribe_events(self, *events: typing.Type[domain_event.MigrationEvent]) -> None:
        ...

    @abc.abstractmethod
    def subscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def unsubscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def get_event_handlers_for(
        self,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> typing.MutableSequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]]:
        ...

    @abc.abstractmethod
    def prioritize_handler(
        self,
        handler: domain_event.EventHandler,
        event: typing.Type[domain_event.MigrationEvent],
        priority: int,
    ) -> None:
        ...

    @abc.abstractmethod
    def unprioritize_handler_proxy(
        self,
        handler_proxy: domain_event.EventHandlerProxy,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        ...

    @abc.abstractmethod
    def listen(
        self,
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> typing.Callable[
        [domain_event.EventHandlerProxyOr[domain_event.EventHandlerT]],
        domain_event.EventHandlerProxyOr[domain_event.EventHandlerT],
    ]:
        ...

    @abc.abstractmethod
    def dispatch(self, event: domain_event.MigrationEvent) -> None:
        ...


class MigrationAppImpl(MigrationApp):
    __slots__: typing.Sequence[str] = (
        "_config",
        "_session",
        "_event_manager",
        "_migration_service",
    )

    def __init__(
        self,
        configuration: config.Config,
    ) -> None:
        ux.init_logging(configuration)
        ux.init_components(configuration)

        self._session = app_session = session.MigrationSessionImpl(self, configuration)
        self._migration_service = migration_service.MigrationService(app_session)

        self._event_manager = event_manager.MigrationEventManagerImpl()
        for event_type, event_handlers in configuration.application.app_subscribed_events.items():
            for handler in event_handlers:
                self._event_manager.subscribe_event_handler(handler, event_type)

        self._event_manager.dispatch(domain_event.StartingEvent(self))

    @property
    def name(self) -> str:
        return self._session.session_name

    @property
    def session(self) -> session.MigrationSession:
        return self._session

    @property
    def event_manager(self) -> domain_event_manager.MigrationEventManager:
        return self._event_manager

    def subscribe_events(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        return self._event_manager.subscribe_events(handler, *events)

    def unsubscribe_events(self, *events: typing.Type[domain_event.MigrationEvent]) -> None:
        return self._event_manager.unsubscribe_events(*events)

    def subscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        return self._event_manager.subscribe_event_handler(handler, event)

    def unsubscribe_event_handler(
        self,
        handler: domain_event.EventHandlerProxyOr[domain_event.EventHandler],
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        return self._event_manager.unsubscribe_event_handler(handler, event)

    def get_event_handlers_for(
        self,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> typing.MutableSequence[domain_event.EventHandlerProxyOr[domain_event.EventHandler]]:
        return self._event_manager.get_event_handlers_for(event)

    def prioritize_handler(
        self,
        handler: domain_event.EventHandler,
        event: typing.Type[domain_event.MigrationEvent],
        priority: int,
    ) -> None:
        return self._event_manager.prioritize_handler(handler, event, priority)

    def unprioritize_handler_proxy(
        self,
        handler_proxy: domain_event.EventHandlerProxy,
        event: typing.Type[domain_event.MigrationEvent],
    ) -> None:
        return self._event_manager.unprioritize_handler_proxy(handler_proxy, event)

    def listen(
        self,
        *events: typing.Type[domain_event.MigrationEvent],
    ) -> typing.Callable[
        [domain_event.EventHandlerProxyOr[domain_event.EventHandlerT]],
        domain_event.EventHandlerProxyOr[domain_event.EventHandlerT],
    ]:
        return self._event_manager.listen(*events)

    def dispatch(self, event: domain_event.MigrationEvent) -> None:
        return self._event_manager.dispatch(event)

    @requires_migrations(is_applied=False)
    def upgrade_once(self) -> int:
        pending_migration_model = self._session.get_migration_model_by_flag(is_applied=False)
        assert pending_migration_model is not None  # Only for type checkers

        pending_migration = self._migration_service.get_migration(
            pending_migration_model.name,
            pending_migration_model.version,
        )

        with self._session.begin_mongo_session() as session_context:
            _LOGGER.info(
                "%s: upgrading waiting migration (#%s -> #%s)...",
                self.name,
                versioning_service.get_previous_migration_version(pending_migration),
                pending_migration.version,
            )

            with self._session.begin_transaction(
                transactions.UpgradeTransaction,
                migration=pending_migration,
            ) as transaction:
                transaction.apply_to(session_context)

                _LOGGER.info(
                    "%s: Successfully upgraded to (#%s).",
                    self.name,
                    pending_migration.version,
                )
                return transactions.TRANSACTION_SUCCESS

    @requires_migrations(is_applied=True)
    def downgrade_once(self) -> int:
        applied_migration_model = self._session.get_migration_model_by_flag(is_applied=True)
        assert applied_migration_model is not None  # Only for type checkers

        applied_migration = self._migration_service.get_migration(
            applied_migration_model.name,
            applied_migration_model.version,
        )

        with self._session.begin_mongo_session() as session_context:
            _LOGGER.info(
                "%s: downgrading waiting migration (#%s -> #%s)...",
                self.name,
                applied_migration.version,
                versioning_service.get_previous_migration_version(applied_migration),
            )

            with self._session.begin_transaction(
                transactions.DowngradeTransaction,
                migration=applied_migration,
            ) as transaction:
                transaction.apply_to(session_context)

                _LOGGER.info(
                    "%s: successfully downgraded to (#%s).",
                    self.name,
                    versioning_service.get_previous_migration_version(applied_migration),
                )
                return transactions.TRANSACTION_SUCCESS

    @requires_migrations(is_applied=False)
    def upgrade_while(
        self, predicate: typing.Callable[[domain_migration.Migration], bool], /
    ) -> int:
        upgraded = 0
        pending_migration_models = self._session.get_migration_models_by_flag(is_applied=False)

        with self._session.begin_mongo_session() as session_context:
            while pending_migration_models:
                migration = self._migration_service.get_migration(
                    (model := pending_migration_models.pop(0)).name,
                    model.version,
                )

                if not predicate(migration):
                    break

                _LOGGER.info(
                    "%s: upgrading waiting migration (#%s -> #%s)...",
                    self.name,
                    versioning_service.get_previous_migration_version(migration),
                    migration.version,
                )

                with self._session.begin_transaction(
                    transactions.UpgradeTransaction,
                    migration=migration,
                ) as transaction:
                    transaction.apply_to(session_context)

                _LOGGER.info(
                    "%s: Successfully upgraded to (#%s).",
                    self.name,
                    migration.version,
                )
                upgraded += 1

            return upgraded

    @requires_migrations(is_applied=True)
    def downgrade_while(
        self, predicate: typing.Callable[[domain_migration.Migration], bool], /
    ) -> int:
        downgraded = 0
        applied_migration_models = self._session.get_migration_models_by_flag(is_applied=True)

        with self._session.begin_mongo_session() as session_context:
            while applied_migration_models:
                migration = self._migration_service.get_migration(
                    (model := applied_migration_models.pop(0)).name,
                    model.version,
                )

                if not predicate(migration):
                    break

                _LOGGER.info(
                    "%s: downgrading waiting migration (#%s -> #%s)...",
                    self.name,
                    migration.version,
                    versioning_service.get_previous_migration_version(migration),
                )

                with self._session.begin_transaction(
                    transactions.DowngradeTransaction,
                    migration=migration,
                ) as transaction:
                    transaction.apply_to(session_context)

                _LOGGER.info(
                    "%s: successfully downgraded to (#%s).",
                    self.name,
                    versioning_service.get_previous_migration_version(migration),
                )
                downgraded += 1

            return downgraded

    def downgrade_to(self, migration_version: int, /) -> int:
        if not migration_version:
            return self.downgrade_all()

        model = self.session.get_migration_model_by_version(migration_version)
        if model is None:
            raise ValueError(f"Migration with version {migration_version!r} is not found.")

        if not model.is_applied:
            raise ValueError(f"Migration with version {migration_version} is already pending.")

        return self.downgrade_while(lambda m: m.version > migration_version)

    def upgrade_to(self, migration_version: int, /) -> int:
        model = self.session.get_migration_model_by_version(migration_version)
        if model is None:
            raise ValueError(f"Migration with version {migration_version!r} is not found.")

        if model.is_applied:
            raise ValueError(f"Migration with version {migration_version} is already applied.")

        return self.upgrade_while(lambda m: m.version <= migration_version)

    def downgrade_all(self) -> int:
        return self.downgrade_while(lambda _: True)

    def upgrade_all(self) -> int:
        return self.upgrade_while(lambda _: True)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = ("MongoModelRepositoryImpl",)

import enum
import threading
import typing

import pymongo

from mongorunway import mongo
from mongorunway.application.ports import repository as repository_port
from mongorunway.domain import migration as domain_migration


class Index(enum.Enum):
    UNAPPLIED = [("is_applied", pymongo.ASCENDING)]
    APPLIED = [("is_applied", pymongo.ASCENDING), ("_id", pymongo.DESCENDING)]
    UNIQUE = "_id_"

    def translate(self) -> str:
        return mongo.translate_index(self.value)


class MongoModelRepositoryImpl(repository_port.MigrationModelRepository):
    __slots__: typing.Sequence[str] = (
        "_collection",
        "_lock",
    )

    def __init__(self, migrations_collection: mongo.Collection) -> None:
        self._collection = migrations_collection
        self._lock = threading.RLock()  # Use reentrant lock to allow nested acquire/release

    def __len__(self) -> int:
        with self._lock:
            return self.has_migrations()

    def __contains__(self, item: typing.Any, /) -> bool:
        with self._lock:
            return self.has_migration(item)

    def has_migration(self, item: typing.Any, /) -> bool:
        version: typing.Optional[int] = getattr(item, "version", None)
        if version is None:
            return NotImplemented

        with self._lock:
            return self.has_migration_with_version(version)

    def has_migration_with_version(self, migration_version: int, /) -> bool:
        with self._lock:
            return self._collection.count_documents(
                {"_id": migration_version}
            ) > 0

    def has_migrations(self) -> bool:
        with self._lock:
            return bool(
                self._collection.count_documents(
                    {},
                    limit=1,
                )
            )

    def acquire_migration_model_by_version(
        self,
        migration_version: int,
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        with self._lock:
            schema = self._collection.find_one({"_id": migration_version})

        if schema is not None:
            return domain_migration.MigrationReadModel.from_dict(schema)

        return None

    def acquire_migration_model_by_flag(
        self, is_applied: bool
    ) -> typing.Optional[domain_migration.MigrationReadModel]:
        with self._lock:
            if is_applied:
                # LIFO
                schema = self._collection.find({"is_applied": True}).sort("_id", -1).limit(1)
            else:
                # FIFO
                schema = self._collection.find({"is_applied": False}).sort("_id", 1).limit(1)

        try:
            model = domain_migration.MigrationReadModel.from_dict(schema.next())
        except StopIteration:
            return None

        return model

    def acquire_migration_models_by_flag(
        self, *, is_applied: bool
    ) -> typing.Iterator[domain_migration.MigrationReadModel]:
        indexes = Index.APPLIED if is_applied else Index.UNAPPLIED
        with self._lock:
            schemas = mongo.hint_or_sort_cursor(
                self._collection.find({"is_applied": is_applied}),
                indexes=indexes.value,
            )

        while True:
            try:
                schema = schemas.next()
            except StopIteration:
                break

            yield domain_migration.MigrationReadModel.from_dict(schema)

    def acquire_all_migration_models(
        self,
        *,
        ascending_id: bool = True,
    ) -> typing.Iterator[domain_migration.MigrationReadModel]:
        with self._lock:
            if ascending_id:
                # By default, the collection has already created an index for the
                # unique key `_id` which sorts them in ascending order.
                schemas = mongo.hint_or_sort_cursor(
                    self._collection.find({}),
                    indexes=Index.UNIQUE.value,
                )

            else:
                schemas = self._collection.find({}).sort([("version", pymongo.DESCENDING)])

        while True:
            try:
                schema = schemas.next()
            except StopIteration:
                break

            yield domain_migration.MigrationReadModel.from_dict(schema)

    def append_migration(self, migration: domain_migration.Migration, /) -> int:
        schema = migration.to_dict(unique=True)

        with self._lock:
            self._collection.insert_one(
                schema,
                bypass_document_validation=True,
            )

        return migration.version

    def remove_migration(self, migration_version: int, /) -> int:
        with self._lock:
            self._collection.delete_one({"_id": migration_version})

        return migration_version

    def set_applied_flag(self, migration: domain_migration.Migration, is_applied: bool) -> int:
        with self._lock:
            self._collection.update_one(
                {"_id": migration.version},
                {"$set": {"is_applied": is_applied}},
                bypass_document_validation=True,
            )

        return migration.version



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "format_auditlog_entry",
    "format_app_date",
)

import datetime
import typing

if typing.TYPE_CHECKING:
    from mongorunway.application import applications
    from mongorunway.domain import migration_auditlog_entry as domain_auditlog_entry


def format_app_date(
    application: applications.MigrationApp,
    date_parts: typing.Optional[typing.Sequence[str]],
) -> typing.Optional[datetime.datetime]:
    if date_parts is None:
        return None

    return datetime.datetime.strptime(
        " ".join(date_parts),
        application.session.session_date_format,
    )


def format_auditlog_entry(
    entry: domain_auditlog_entry.MigrationAuditlogEntry, /
) -> typing.Sequence[str]:
    migration = entry.migration_read_model

    failed_msg = "False"
    if entry.is_failed():
        failed_msg = f"Error: {entry.exc_name}\nMessage: {entry.exc_message}"

    return [
        entry.format_date(),
        failed_msg,
        entry.transaction_name,
        f"Name: {migration.name}\n"
        f"Version: {migration.version}\n"
        f"Is applied: {migration.is_applied}\n",
    ]



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "pass_application",
    "auditlog",
    "create_template",
    "cli",
    "status",
    "walk",
    "downgrade",
    "upgrade",
    "refresh",
    "version",
    "init",
)

import functools
import sys
import typing

import click

from mongorunway.application import applications
from mongorunway.application import use_cases
from mongorunway.presentation import presenters

_P = typing.ParamSpec("_P")
_T = typing.TypeVar("_T")
_CommandT = typing.TypeVar("_CommandT")


def pass_application(command: typing.Callable[_P, _T]) -> typing.Callable[_P, _T]:
    click.option("--config-file", type=click.STRING)(command)

    @click.pass_context
    def wrapper(ctx: click.Context, *args: _P.args, **kwargs: _P.kwargs) -> _T:
        configuration = use_cases.read_configuration(
            ctx.params["config_file"],
            app_name=ctx.params["application_name"],
            verbose_exc=ctx.params.get("verbose_exc", False),
        )

        if configuration is use_cases.UseCaseFailed:
            ctx.fail("Configuration failed.")

        application = applications.MigrationAppImpl(configuration)
        return typing.cast(_T, ctx.invoke(command, *args, application=application, **kwargs))

    return typing.cast(
        typing.Callable[_P, _T],
        functools.update_wrapper(wrapper, command),
    )


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.option(
    "--verbose",
    is_flag=True,
    help="Enable verbose output during the upgrade process."
)
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the upgrade process."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@click.argument(
    "expression",
    type=click.STRING,
    default="+",
    metavar="EXPRESSION",
    required=False,
)
@pass_application
def upgrade(
    application: applications.MigrationApp,
    expression: str,
    verbose: bool,
    verbose_exc: bool,
    **params: typing.Any,
) -> None:
    """Upgrade the specified application.

    This command allows you to upgrade the specified application using the
    given expression or additional arguments. Optionally, you can enable
    verbose output or verbose output for exceptions during the upgrade process.
    """

    exit_code = use_cases.upgrade(
        application=application,
        expression=expression,
        verbose=verbose,
        verbose_exc=verbose_exc,
    )
    sys.exit(exit_code)


@cli.command()
@click.option(
    "--verbose",
    is_flag=True,
    help="Enable verbose output during the downgrade process."
)
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the downgrade process."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@click.argument(
    "expression",
    type=click.STRING,
    default="-",
    metavar="EXPRESSION",
    required=False,
)
@pass_application
def downgrade(
    application: applications.MigrationApp,
    expression: str,
    verbose: bool,
    verbose_exc: bool,
    **params: typing.Any,
) -> None:
    """Downgrade the specified application.

    This command allows you to downgrade the specified application using the
    given expression or additional arguments. Optionally, you can enable verbose
    output or verbose output for exceptions during the downgrade process.
    """

    exit_code = use_cases.downgrade(
        application=application,
        expression=expression,
        verbose=verbose,
        verbose_exc=verbose_exc,
    )
    sys.exit(exit_code)


@cli.command()
@click.option(
    "--verbose",
    is_flag=True,
    help="Enable verbose output during the walk process."
)
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the walk process."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@click.argument(
    "expression",
    type=click.STRING,
    metavar="EXPRESSION",
    required=True,
)
@pass_application
def walk(
    application: applications.MigrationApp,
    expression: str,
    verbose: bool,
    verbose_exc: bool,
    **params: typing.Any,
) -> None:
    """
    Walk through the specified application.

    This command allows you to perform a walk operation on the specified
    application using the given expression or additional arguments. The
    walk operation enables you to traverse the application's structure or
    perform specific actions.

    Optionally, you can enable verbose output or verbose output for exceptions
    during the walk process.
    """

    exit_code = use_cases.walk(
        application=application,
        expression=expression,
        verbose=verbose,
        verbose_exc=verbose_exc,
    )
    sys.exit(exit_code)


@cli.command()
@click.option(
    "-n",
    "--name",
    type=click.STRING,
    help="Specify the name of the migration template."
)
@click.option(
    "-v",
    "--version",
    type=click.INT,
    default=None,
    required=False,
    help="Specify the version number of the migration template."
)
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the template creation process."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@pass_application
def create_template(
    application: applications.MigrationApp,
    name: str,
    verbose_exc: bool,
    version: typing.Optional[int] = None,
    **params: typing.Any,
) -> None:
    """Create a migration template for the specified application.

    This command allows you to create a migration template for the specified
    application. The template can be customized with a name and an optional
    version number.

    Optionally, you can enable verbose output for exceptions during the
    template creation process.
    """

    exit_code = use_cases.create_migration_file(
        application=application,
        migration_filename=name,
        migration_version=version,
        verbose_exc=verbose_exc,
    )
    sys.exit(exit_code)


@cli.command()
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose exception mode.",
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
def refresh(
    application: applications.MigrationApp,
    verbose_exc: bool,
) -> None:
    """Refreshes the specified application.

    This command refreshes the specified application by performing certain actions.
    It accepts an APPLICATION_NAME as a required argument and provides an option
    to enable verbose exception mode using the --verbose-exc flag.
    """
    exit_code = use_cases.refresh(
        application=application,
        verbose_exc=verbose_exc,
    )
    sys.exit(exit_code)


@cli.command()
@click.option(
    "--verbose",
    is_flag=True,
    help="Enable verbose output for the status command."
)
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the status command."
)
@click.option(
    "-d",
    "--depth",
    type=click.INT,
    default=-1,
    required=False,
    help="Specify the depth of the migration history to display."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@pass_application
def status(
    application: applications.MigrationApp,
    depth: int,
    verbose: bool,
    verbose_exc: bool,
    **params: typing.Any,
) -> None:
    """Display the migration status for the specified application.

    This command allows you to view the migration status of the specified
    application. It shows the history of applied migrations up to the
    specified depth.

    Optionally, you can enable verbose output for the status command and
    verbose output for exceptions that may occur during the status check.
    """

    presenters.show_status(
        application=application,
        pushed_depth=depth,
        verbose=verbose,
        verbose_exc=verbose_exc,
    )


@cli.command()
@click.option(
    "-s",
    "--start",
    type=click.STRING,
    default=None,
    nargs="?",
    help="Specify the start timestamp or date for filtering audit log entries."
)
@click.option(
    "-e",
    "--end",
    type=click.STRING,
    default=None,
    nargs="?",
    help="Specify the end timestamp or date for filtering audit log entries."
)
@click.option(
    "-l",
    "--limit",
    type=click.INT,
    default=10,
    help="Specify the maximum number of audit log entries to display."
)
@click.option(
    "-a",
    "--ascending",
    is_flag=True,
    help="Sort the audit log entries in ascending order by date."
)
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the audit log command."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@pass_application
def auditlog(
    application: applications.MigrationApp,
    start: typing.Optional[typing.Sequence[str]],
    end: typing.Optional[typing.Sequence[str]],
    limit: int,
    ascending: bool,
    verbose_exc: bool,
    **params: typing.Any,
) -> None:
    """Display the audit log entries for the specified application.

    This command allows you to view the audit log entries for the
    specified application. You can filter the entries by specifying a
    start and/or end timestamp or date. The maximum number of entries
    displayed can be limited, and you can choose to sort the entries
    in ascending order by date.

    Optionally, you can enable verbose output for exceptions that may
    occur during the audit log command.
    """

    presenters.show_auditlog_entries(
        application=application,
        start=start,
        end=end,
        limit=limit,
        verbose_exc=verbose_exc,
        ascending_date=ascending,
    )


@cli.command()
@click.option(
    "--verbose",
    is_flag=True,
    help="Enable verbose output for the version command."
)
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the version command."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@pass_application
def version(application: applications.MigrationApp, verbose: bool, **params: typing.Any) -> None:
    """Display the version information for the specified application.

    This command allows you to view the version information of the
    specified application. It shows details such as the application's
    name, version number, and other relevant details.

    Optionally, you can enable verbose output for the version command
    to get more detailed information.
    """

    presenters.show_version(application=application, verbose=verbose)


@cli.command()
@click.option(
    "--verbose-exc",
    is_flag=True,
    help="Enable verbose output for exceptions during the init command."
)
@click.option(
    "--scripts-dir",
    is_flag=True,
    help="Initialize the scripts directory for the specified application."
)
@click.option(
    "--collection",
    is_flag=True,
    help="Initialize the collection for the specified application."
)
@click.option(
    "--indexes",
    is_flag=True,
    help="Initialize the indexes for the specified application's collection."
)
@click.option(
    "--schema-validation",
    is_flag=True,
    help="Enable schema validation for the specified application's collection."
)
@click.argument(
    "application_name",
    type=click.STRING,
    metavar="APPLICATION_NAME",
    required=True,
)
@pass_application
def init(
    application: applications.MigrationApp,
    scripts_dir: bool,
    collection: bool,
    indexes: bool,
    verbose_exc: bool,
    schema_validation: bool,
    **params: typing.Any,
) -> None:
    """Initialize the specified application for migration.

    This command allows you to initialize the specified application
    for migration. You can choose to initialize the scripts directory,
    collection, indexes, and enable schema validation for the
    application's collection.

    Initializing the scripts directory will create the necessary
    directory structure for storing migration scripts.

    Initializing the collection will create the migration collection in
    the database.

    Initializing the indexes will create the required indexes for the
    migration collection.

    Enabling schema validation will enforce the defined schema on the
    migration collection.

    Note: Use caution when running the init command as it may modify
    the application's database.

    Optionally, you can enable verbose output for exceptions that may
    occur during the init command.
    """

    exit_code = use_cases.init(
        application=application,
        verbose_exc=verbose_exc,
        init_scripts_dir=scripts_dir,
        init_collection=collection,
        init_collection_indexes=indexes,
        init_collection_schema_validation=schema_validation,
    )
    sys.exit(exit_code)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
r"""Utilities.

The module contains utilities that can be used in a project.

!!! info
    This module is independent of all other modules and can only use
    built-in or third-party libraries.
"""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "get_module",
    "as_snake_case",
    "import_obj",
    "convert_string",
    "build_mapping_values",
    "build_optional_kwargs",
    "is_valid_filename",
    "hexlify",
)

import binascii
import copy
import distutils.util
import importlib
import importlib.util
import os
import re
import time
import types
import typing

import bson
import typing_extensions

try:
    _P = typing.ParamSpec("_P")
except AttributeError:
    _P = typing_extensions.ParamSpec("_P")

_T = typing.TypeVar("_T")
_TT = typing.TypeVar("_TT", bound=typing.Type[typing.Any])

string_case_pattern: typing.Final[typing.Pattern[str]] = re.compile(
    r"""
    (?<=_)     # Positive lookbehind assertion for an underscore character
    (\w+)      # Match and capture one or more word characters (alphanumeric or underscore)
    |          # Or
    ([A-Z][a-z]+)   # Match and capture an uppercase letter followed by one or more lowercase letters
    |          # Or
    ([a-z]+)   # Match and capture one or more lowercase letters
    (?=[A-Z])  # Positive lookahead assertion for an uppercase letter
    """,
    flags=re.VERBOSE,
)
r"""Regular expression for string cases.

The `string_case_pattern` constant is a regular expression pattern 
used for matching different cases in strings.
"""

number_pattern: typing.Final[typing.Pattern[str]] = re.compile(
    r"""
    [-+]?                    # Matches an optional sign character (- or +)
    (?:                       # Non-capturing group:
       \d{1,3}                # Matches 1 to 3 digits
       (?:_\d{3})*            # Matches an underscore (_) followed by exactly 3 digits, 
                              #                             repeated zero or more times.
       |                      # Alternation operator, allows matching either the previous pattern 
                              #                                         or the following pattern.
       [\d_]+                 # Matches one or more digits or underscores
       (?:\.\d+)?             # Matches an optional decimal part, where \.\d+ matches a dot 
                              #                                 followed by one or more digits.
    )
    """,
    flags=re.VERBOSE,
)
r"""Regular expression for numeric values.

The number_pattern constant is a regular expression pattern used for 
matching numeric values.
"""


@typing.overload
def convert_string(val: typing.Literal["true", "yes", "ok"], /) -> typing.Literal[True]:
    ...


@typing.overload
def convert_string(val: typing.Literal["false", "no"], /) -> typing.Literal[False]:
    ...


@typing.overload
def convert_string(val: typing.Literal["none", "nothing", "undefined"], /) -> typing.Literal[None]:
    ...


def convert_string(value: str, /) -> typing.Union[int, float, bool, str, None]:
    r"""Converts a string.

    Converts a string depending on the received value. Case-insensitive
    string conversion. If the string fails all the checks, the passed
    value will be returned.

    Parameters
    ----------
    value : str
        The value that needs to be converted.

    Returns
    -------
    typing.Union[int, float, bool, str, None]
        None, if the value is one of {"none", "nothing", "undefined"}.

    See Also
    --------
    number_pattern : Relationship.
    distutils.util.strtobool : Relationship.
    """
    if not isinstance(value, str):
        return value  # type: ignore[unreachable]

    value_lower = value.lower().strip()
    if value_lower in {"none", "nothing", "undefined"}:
        return None

    if digit_matches := re.findall(number_pattern, value_lower):
        # For floating point numbers.
        if len(digit_matches) > 1:
            # Floating-point numbers should not have more than one decimal point.
            digit_matches = digit_matches[:2]

            # To handle numbers that use an underscore.
            if any(m.count("_") for m in digit_matches):
                return int("".join(digit_matches))

            return float(".".join(digit_matches))

        (digit_string,) = digit_matches
        return int(digit_string)

    try:
        integer = distutils.util.strtobool(value_lower)
    except ValueError:
        return value

    return bool(integer)


def hexlify(binary: bson.binary.Binary) -> str:
    r"""Converts a binary object to hexadecimal.

    Converts a binary object to a more user-friendly hexadecimal string.
    This method can be useful for converting session IDs in both pymongo
    and mongorunway, as these tools use the same format for identifiers.

    Parameters
    ----------
    binary : bson.binary.Binary
        Binary representation of the object to be converted to a hexadecimal
        string.

    Returns
    -------
    str
        The converted binary object as a hexadecimal string.
    """
    hex_id = binascii.hexlify(binary).decode()
    return hex_id


def build_mapping_values(
    mapping: typing.MutableMapping[_T, typing.Any], /
) -> typing.MutableMapping[_T, typing.Any]:
    r"""Converts string values in a mutable mapping.

    Converts string values in a mutable mapping to corresponding named
    objects. If a string fails validation, the original value is retained.

    Parameters
    ----------
    mapping : typing.MutableMapping[_T, typing.Any]
        Mutable mapping whose values need to be converted.

    Returns
    -------
    typing.MutableMapping[_T, typing.Any]
        Mutable mapping with converted values.

    Example
    -------
    ```py
    >>> build_mapping_values({1: "undefined", 2: "ok", 3: "no"})
    {1: None, 2: True, 3: False}
    ```

    See Also
    --------
    convert_string
    """
    for key, value in copy.copy(mapping).items():
        mapping[key] = convert_string(value)

    return mapping


def build_optional_kwargs(
    keys: typing.Iterable[_T],
    mapping: typing.MutableMapping[typing.Any, typing.Any],
) -> typing.MutableMapping[_T, typing.Any]:
    r"""Converts string values in a mutable mapping.

    Converts string values in a mutable mapping to corresponding named
    objects. If the value of a dictionary is None, the loop will move
    to the next iteration. Returns a mutable mapping with the keys passed
    to this function.

    Parameters
    ----------
    keys : typing.Iterable[_T]
        Keys of optional values to be converted if their value is not None.
    mapping : typing.MutableMapping[typing.Any, typing.Any]
        Mutable mapping whose values need to be converted.

    Returns
    -------
    typing.MutableMapping[_T, typing.Any]
        Mutable mapping with all non-None values converted.

    Example
    -------
    ```py
    >>> build_optional_kwargs([1, 2], {1: None, 2: "ok", 3: "no"})
    {2: True}
    ```

    See Also
    --------
    convert_string
    """
    kwargs = {}

    for key in keys:
        value = mapping.get(key)
        if value is not None:
            kwargs[key] = convert_string(value)

    return kwargs


def get_module(directory: str, filename: str) -> types.ModuleType:
    r"""Loads a module from given directory.

    Load and return a module based on the given directory and filename.
    Also supports the format without a file extension e.g. '.py' .

    Parameters
    ----------
    directory : str
        The path to the directory where the file is located.
    filename : str
        The name of the file containing the module.

    Returns
    -------
    types.ModuleType
        The module object that is loaded from the file.

    Raises
    ------
    ModuleNotFoundError
        If the file specified by `directory` and `filename` parameters
        cannot be found.

    Example
    -------
    ```py
    >>> from types import ModuleType
    ... # Obtains the current file util.py when running doctest
    ... # in the current directory.
    >>> assert isinstance(get_module("", "util.py"), ModuleType)
    ...
    >>> from types import ModuleType
    ... # Also supports the format without a file extension.
    >>> assert isinstance(get_module("", "util"), ModuleType)
    ```

    See Also
    --------
    importlib.spec_from_file_location
    """
    if not filename.endswith(".py"):
        filename += ".py"

    spec = importlib.util.spec_from_file_location(
        filename.rstrip(".py"), (path := os.path.join(directory, filename))
    )
    if spec is None:
        raise ModuleNotFoundError(f"Module {path!r} is not found.")

    module = importlib.util.module_from_spec(spec)
    assert spec.loader is not None  # For type checkers only
    spec.loader.exec_module(module)
    return module


def import_obj(obj_path: str, /, cast: _TT) -> _TT:
    r"""Imports a class from the specified module.

    Imports a class from the specified module. The module path should
    be specified using dot notation, with the class itself at the end.

    Parameters
    ----------
    obj_path : str
        The path to the obj to be imported.
    cast : _TT
        The type to which the value should be cast. This is used to
        provide more clarity in the code. This parameter is also useful
        for compatibility with type checkers.

    Returns
    -------
    _TT
        The type to which we cast the imported obj.

    See Also
    --------
    importlib.import_module
    """
    module_name, obj_name = obj_path.rsplit(".", maxsplit=1)
    module = importlib.import_module(module_name)
    return typing.cast(cast, getattr(module, obj_name))


def is_valid_filename(directory: str, filename: str) -> bool:
    r"""Validates the file in the given directory.

    Check whether a given filename is a valid migration file in the
    given directory.

    Parameters
    ----------
    directory : str
        The path of the directory where the file is located.
    filename : str
        The name of the file to be checked.

    Returns
    -------
    bool
        True if the filename is a valid migration file, False otherwise.

    Raises
    ------
    ValueError
        If the value passed to the directory parameter is not a directory.

    Notes
    -----
    The function checks if a file exists in the specified directory.
    Additionally, the file must have the extension '.py' and should not
    start with a dunder[^1].

    References
    ----------
    [^1]:
        Dunder: The term "dunder" is a short form of "double underscore"
        and is commonly used to refer to special methods or attributes in
        Python that are surrounded by double underscores on both sides, such
        as __init__ or __name__.
    """
    if not os.path.isdir(directory):
        raise ValueError(f"The specified path {directory!r} is not a directory.")

    return (
        os.path.isfile(os.path.join(directory, filename))
        and filename.endswith(".py")
        and not filename.startswith("__")
    )


def as_snake_case(obj: typing.Any) -> str:
    r"""Converts an object to snake case format.

    Converts an object to snake case format regardless of the type of the
    object passed. If the object is an instance of the `str` class, the object
    itself will be converted. Otherwise, the `obj.__name__` will be taken,
    and if it is not available, the attribute will be taken from the type of
    the object passed.

    Parameters
    ----------
    obj : typing.Any
        The object, name, or value to be converted to snake case.

    Returns
    -------
    str
        The converted name of the object or its value in snake case. It may
        return the original object if the object does not match the regular
        expression.

    Example
    -------
    ```py
    >>> class DummyClass: pass
    >>> dummy_obj = DummyClass()
    >>> as_snake_case(DummyClass)
    'dummy_class'
    >>> as_snake_case(dummy_obj)  # Takes `type(dummy_obj).__name__` .
    'dummy_class'
    >>> as_snake_case("snake_case_str")
    'snake_case_str'
    >>> as_snake_case("camelCase")
    'camel_case'
    ```
    """
    if not isinstance(obj, str):
        obj_name = getattr(obj, "__name__", None)
        obj = type(obj).__name__ if obj_name is None else obj_name

    if obj.count("_"):
        # Does not support mix of string cases and skips snake case.
        return obj

    obj_match = re.findall(string_case_pattern, obj)
    obj_name = "_".join(("".join(i.lower() for i in g) for g in obj_match))

    return obj_name


def timeit_func(
    func: typing.Callable[_P, _T],
    *args: _P.args,
    **kwargs: _P.kwargs,
) -> typing.Tuple[_T, float]:
    r"""Benchmarks the execution time of a function.

    Returns the execution time of a function and the resulting output of
    calling the function with given parameters.

    Parameters
    ----------
    func : Callable[_P, _T]
        The function whose execution time needs to be benched.
    args : _P.args
        Positional arguments of a function.
    kwargs : _P.kwargs
        Keyword arguments of a function

    Returns
    -------
    Tuple[_T, float]
        A tuple where the first element is the type of the returned result
        of the function, and the second element is the time taken to execute
        the function.

    See Also
    --------
    SystemTimer
    """
    with SystemTimer() as timer:
        result = func(*args, **kwargs)

    return result, timer.executed_in


class SystemTimer:
    """Implementation of a system timer for code benchmarks.

    This class represents an implementation of a system timer for measuring
    the execution time of a specific code block. It is a context manager and
    can be conveniently used with the `with` statement.
    """

    __slots__: typing.Sequence[str] = (
        "_start",
        "_executed_in",
    )

    def __init__(self) -> None:
        self._start = 0.0
        self._executed_in = 0.0

    @property
    def start(self) -> float:
        """Returns the start time of the benchmark.

        Returns the system time that was set in the `__enter__` method of
        this class.

        Returns
        -------
        float
            The system time when the code block started execution.
        """
        return self._start

    @property
    def executed_in(self) -> float:
        """Returns the execution time of a code block.

        Returns the time it took for the code block to execute. The
        calculation is done using the system time from the `time` module.

        Returns
        -------
        float
            The execution time of the code block.
        """
        return self._executed_in

    def __enter__(self) -> SystemTimer:
        self._start = time.time()
        return self

    @typing.overload
    def __exit__(self, exc_type: None, exc_val: None, exc_tb: None) -> None:
        ...

    @typing.overload
    def __exit__(
        self,
        exc_type: typing.Type[BaseException],
        exc_val: BaseException,
        exc_tb: types.TracebackType,
    ) -> None:
        ...

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_val: typing.Optional[BaseException],
        exc_tb: typing.Optional[types.TracebackType],
    ) -> None:
        end = time.time()
        self._executed_in = end - self.start



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MIGRATION_SCHEMA_VALIDATOR",
    "PENDING_MIGRATION_INDEX",
    "APPLIED_MIGRATION_INDEX",
    "ValidationAction",
    "ValidationLevel",
    "configure_logging",
    "configure_migration_indexes",
    "configure_migration_directory",
    "configure_migration_collection",
    "configure_migration_schema_validators",
    "init_components",
    "init_logging",
    "init_migration_indexes",
    "init_schema_validators",
    "init_migration_collection",
    "init_migration_directory",
    "remove_migration_indexes",
    "remove_migration_schema_validators",
    "sync_scripts_with_repository",
)

import enum
import logging
import logging.config
import os
import typing

from mongorunway.application.services import migration_service

if typing.TYPE_CHECKING:
    from mongorunway import mongo
    from mongorunway.application import config
    from mongorunway.application import traits

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("mongorunway.ux")

MIGRATION_SCHEMA_VALIDATOR: typing.Final[typing.Dict[str, typing.Any]] = {
    "$jsonSchema": {
        "bsonType": "object",
        "required": [
            "_id",
            "name",
            "version",
            "checksum",
            "is_applied",
            "description",
        ],
        "properties": {
            "_id": {
                "bsonType": "int",
            },
            "name": {
                "bsonType": "string",
                "minLength": 1,
            },
            "version": {
                "bsonType": "int",
                "minimum": 1,
            },
            "checksum": {
                "bsonType": "string",
            },
            "is_applied": {
                "bsonType": "bool",
            },
            "description": {
                "bsonType": "string",
            },
        },
    }
}

APPLIED_MIGRATION_INDEX: typing.Final[typing.Sequence[typing.Tuple[str, int]]] = [
    ("is_applied", 1),
    ("_id", -1),
]

PENDING_MIGRATION_INDEX: typing.Final[typing.Sequence[typing.Tuple[str, int]]] = [
    ("is_applied", 1),
]


def configure_logging(config_dict: typing.Dict[str, typing.Any]) -> None:
    if _LOGGER.isEnabledFor(logging.INFO):
        _LOGGER.info("Mongorunway loggers are already configured, skipping...")
    else:
        logging.config.dictConfig(config_dict)
        _LOGGER.info("Mongorunway loggers successfully configured.")


def configure_migration_directory(scripts_dir: str) -> None:
    _LOGGER.info("Checking if the migration directory exists...")
    if not os.path.exists(scripts_dir):
        _LOGGER.info("The migration directory was not found, resolving...")
        os.mkdir(scripts_dir)
        _LOGGER.info(
            "The migration directory has been successfully created at %s",
            scripts_dir,
        )
    else:
        _LOGGER.info("Migration directory is already exists, skipping...")


def configure_migration_collection(
    database: mongo.Database,
    use_schema_validation: bool,
    collection_name: str = "migrations",
) -> None:
    if collection_name not in database.list_collection_names():
        _LOGGER.info("Collection %s is not found, resolving...", collection_name)

        kwargs: typing.Dict[str, typing.Any] = {}
        if use_schema_validation:
            _LOGGER.info(
                "Applying a validator to %s collection...",
                collection_name,
            )
            kwargs["validator"] = MIGRATION_SCHEMA_VALIDATOR

        database.create_collection(
            collection_name,
            **kwargs,
        )

        _LOGGER.info("Migrations collection successfully configured.")
    else:
        _LOGGER.info("Migration collection is already exists, skipping...")


def configure_migration_indexes(collection: mongo.Collection) -> None:
    _LOGGER.info("The 'use_indexes' parameter is set to True, checking for missing indexes...")

    indexes = collection.index_information()

    def _create_index_if_not_exists(index: typing.Sequence[typing.Tuple[str, int]]) -> None:
        translated_index = "_".join(f"{x}_{y}" for x, y in [_ for _ in index])

        if translated_index not in indexes:
            _LOGGER.info(
                "Found one missing index: %s, resolving...",
                translated_index,
            )
            collection.create_index(index)
            _LOGGER.info("Index %s successfully configured.", translated_index)
        else:
            _LOGGER.info(
                "Index %s is already configured, skipping...",
                translated_index,
            )

    _create_index_if_not_exists(APPLIED_MIGRATION_INDEX)
    _create_index_if_not_exists(PENDING_MIGRATION_INDEX)


def remove_migration_indexes(collection: mongo.Collection) -> None:
    _LOGGER.info(
        "The 'use_indexes' parameter is set to False, checking the indexes that "
        "need to be deleted...",
    )

    indexes = collection.index_information()

    def _drop_index_if_exists(index: typing.Sequence[typing.Tuple[str, int]]) -> None:
        translated_index = mongo.translate_index(index)

        if translated_index in indexes:
            _LOGGER.info(
                "Found one existing index: %s, dropping...",
                translated_index,
            )
            collection.drop_index(index)
            _LOGGER.info("Index %s successfully dropped.", translated_index)
        else:
            _LOGGER.info(
                "Index %s is already dropped, skipping...",
                translated_index,
            )

    _drop_index_if_exists(APPLIED_MIGRATION_INDEX)
    _drop_index_if_exists(PENDING_MIGRATION_INDEX)


def configure_migration_schema_validators(collection: mongo.Collection) -> None:
    validator = collection.options().get("validator")
    _LOGGER.info("Schema validation is enabled, checking for validators...")

    if validator != MIGRATION_SCHEMA_VALIDATOR:
        _LOGGER.info("Undefined validator found, removing...")

    collection.database.command(
        "collMod",
        collection.name,
        validationLevel=ValidationLevel.STRICT,
        validationAction=ValidationAction.ERROR,
        validator=MIGRATION_SCHEMA_VALIDATOR,
    )

    _LOGGER.info("Mongorunway migrations schema validator successfully configured.")


def remove_migration_schema_validators(collection: mongo.Collection) -> None:
    validator = collection.options().get("validator")
    _LOGGER.info("Schema validation is disabled, checking for validators...")
    if validator == MIGRATION_SCHEMA_VALIDATOR:
        collection.database.command(
            "collMod",
            collection.name,
            validator={},
        )
        _LOGGER.info("Mongorunway schema validator successfully removed.")


# HIGH-LEVEL UX TOOLS


def sync_scripts_with_repository(
    application: traits.MigrationSessionAware,
) -> typing.Sequence[str]:
    synced = []
    service = migration_service.MigrationService(application.session)
    for migration in service.get_migrations():
        if not application.session.has_migration(migration):
            application.session.append_migration(migration)
            synced.append(migration.name)

            _LOGGER.info(
                "%s: migration '%s' with version %s was synced"
                " "
                "and successfully append to pending.",
                sync_scripts_with_repository.__name__,
                migration.name,
                migration.version,
            )

    return synced


def init_logging(configuration: config.Config, /) -> None:
    if configuration.application.is_logged:
        configure_logging(configuration.logging_dict)


def init_components(configuration: config.Config, /) -> None:
    journal = configuration.application.app_auditlog_journal
    if journal is not None:
        journal.set_max_records(configuration.application.app_auditlog_limit)


def init_migration_directory(configuration: config.Config, /) -> None:
    configure_migration_directory(configuration.filesystem.scripts_dir)


def init_migration_collection(configuration: config.Config, /) -> None:
    configure_migration_collection(
        database=configuration.application.app_database,
        use_schema_validation=configuration.application.use_schema_validation,
    )


def init_migration_indexes(configuration: config.Config, collection: mongo.Collection) -> None:
    if configuration.application.use_indexing:
        configure_migration_indexes(collection)
    else:
        remove_migration_indexes(collection)


def init_schema_validators(configuration: config.Config, collection: mongo.Collection) -> None:
    if configuration.application.use_schema_validation:
        configure_migration_schema_validators(collection)
    else:
        remove_migration_schema_validators(collection)


class ValidationAction(str, enum.Enum):
    ERROR = "error"
    WARNING = "warn"


class ValidationLevel(str, enum.Enum):
    OFF = "off"
    STRICT = "strict"
    MODERATE = "moderate"



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = ("check_if_all_pushed_successfully",)

import typing

from mongorunway.application.services import migration_service

if typing.TYPE_CHECKING:
    from mongorunway.application import applications


def check_if_all_pushed_successfully(
    application: applications.MigrationApp,
    *,
    depth: int = -1,
) -> bool:
    service = migration_service.MigrationService(application.session)

    directory_state = service.get_migrations()
    if not directory_state:
        raise ValueError("Migration files does not exist.")

    applied_state = application.session.get_migration_models_by_flag(is_applied=True)
    if not applied_state:
        raise ValueError("There are currently no applied migrations.")

    if depth > 0:
        if depth > (dir_length := len(directory_state)):
            raise ValueError(
                f"Depth ({depth}) cannot be more than migration files count ({dir_length})."
            )

        return len(directory_state[:depth]) == len(applied_state[:depth])

    return len(directory_state) == len(applied_state)



# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "EventHandler",
    "EventHandlerProxy",
    "EventHandlerProxyOr",
    "EventHandlerT",
    "StartingEvent",
    "ApplicationEvent",
    "MigrationEvent",
)

import typing

import attr

if typing.TYPE_CHECKING:
    from mongorunway.application import applications

EventHandler: typing.TypeAlias = typing.Callable[["MigrationEvent"], None]

EventHandlerT = typing.TypeVar("EventHandlerT", bound=EventHandler)

EventHandlerProxyOr: typing.TypeAlias = typing.Union[EventHandlerT, "EventHandlerProxy"]


@attr.define(eq=True, order=True, hash=True)
class EventHandlerProxy:
    _priority: int = attr.field(
        eq=True,
        hash=True,
        alias="priority",
    )  # type: ignore[call-overload]

    _handler: EventHandler = attr.field(
        eq=False,
        hash=False,
        alias="handler",
    )  # type: ignore[call-overload]

    @property
    def priority(self) -> int:
        return self._priority

    @property
    def handler(self) -> EventHandler:
        return self._handler

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._handler(*args, **kwargs)


@attr.define
class MigrationEvent:
    pass


@attr.define
class ApplicationEvent(MigrationEvent):
    application: applications.MigrationApp = attr.field()


@attr.define
class StartingEvent(ApplicationEvent):
    pass
